@@ -253,6 +253,7 @@ class HeaderResolver(object):
         return self._create_columns("image")
 
     def _create_columns(self, klass):
+        target_class = self.target_object.__class__
         if self.types is not None and len(self.types) != len(self.headers):
             message = "Number of columns and column types not equal."
             raise MetadataError(message)
@@ -308,7 +309,7 @@ class HeaderResolver(object):
                               self.DEFAULT_COLUMN_SIZE, list()))
                 # Ensure ImageColumn is named "Image"
                 column.name = "Image"
-            if column.__class__ is RoiColumn:
+            if column.__class__ is RoiColumn and target_class != DatasetI:
                 append.append(StringColumn(ROI_NAME_COLUMN, '',
                               self.DEFAULT_COLUMN_SIZE, list()))
                 # Ensure RoiColumn is named 'Roi'
@@ -446,7 +447,7 @@ class ValueResolver(object):
             try:
                 return images_by_id[int(value)].id.val
             except KeyError:
-                log.debug('Image Id: %i not found!' % (value))
+                log.debug('Image Id: %s not found!' % (value))
                 return -1
             return
         if WellColumn is column_class:
@@ -458,6 +459,8 @@ class ValueResolver(object):
             return self.wrapper.resolve_dataset(column, row, value)
         if RoiColumn is column_class:
             return self.wrapper.resolve_roi(column, row, value)
+        if column_as_lower == 'shape':
+            return self.wrapper.resolve_shape(value)
         if column_as_lower in ('row', 'column') \
            and column_class is LongColumn:
             try:
@@ -757,8 +760,36 @@ class DatasetWrapper(PDIWrapper):
         super(DatasetWrapper, self).__init__(value_resolver)
         self.images_by_id = dict()
         self.images_by_name = dict()
+        self.rois_by_id = None
+        self.shapes_by_id = None
         self._load()
 
+    def resolve_roi(self, column, row, value):
+        # Support Dataset table with known ROI IDs
+        if self.rois_by_id is None:
+            self._load_rois()
+        try:
+            return self.rois_by_id[int(value)].id.val
+        except KeyError:
+            log.warn('Dataset is missing ROI: %s' % value)
+            return -1
+        except ValueError:
+            log.warn('Wrong input type for ROI ID: %s' % value)
+            return -1
+
+    def resolve_shape(self, value):
+        # Support Dataset table with known Shape IDs
+        if self.rois_by_id is None:
+            self._load_rois()
+        try:
+            return self.shapes_by_id[int(value)].id.val
+        except KeyError:
+            log.warn('Dataset is missing Shape: %s' % value)
+            return -1
+        except ValueError:
+            log.warn('Wrong input type for Shape ID: %s' % value)
+            return -1
+
     def get_image_id_by_name(self, iname, dname=None):
         return self.images_by_name[iname].id.val
 
@@ -800,12 +831,48 @@ class DatasetWrapper(PDIWrapper):
             images_by_id[iid] = image
             if iname in self.images_by_name:
                 raise Exception("Image named %s(id=%d) present. (id=%s)" % (
-                    iname, self.images_by_name[iname], iid
+                    iname, self.images_by_name[iname].id.val, iid
                 ))
             self.images_by_name[iname] = image
         self.images_by_id[self.target_object.id.val] = images_by_id
         log.debug('Completed parsing dataset: %s' % self.target_name)
 
+    def _load_rois(self):
+        log.debug('Loading ROIs in Dataset:%d' % self.target_object.id.val)
+        self.rois_by_id = {}
+        self.shapes_by_id = {}
+        query_service = self.client.getSession().getQueryService()
+        parameters = omero.sys.ParametersI()
+        parameters.addId(self.target_object.id.val)
+        data = list()
+        while True:
+            parameters.page(len(data), 1000)
+            rv = unwrap(query_service.projection((
+                'select distinct i, r, s '
+                'from Shape s '
+                'join s.roi as r '
+                'join r.image as i '
+                'join i.datasetLinks as dil '
+                'join dil.parent as d '
+                'where d.id = :id order by s.id desc'),
+                parameters, {'omero.group': '-1'}))
+            if len(rv) == 0:
+                break
+            else:
+                data.extend(rv)
+        if not data:
+            raise MetadataError("No ROIs on images in target Dataset")
+
+        for image, roi, shape in data:
+            # we only care about *IDs* of ROIs and Shapes in the Dataset
+            rid = roi.id.val
+            sid = shape.id.val
+            self.rois_by_id[rid] = roi
+            self.shapes_by_id[sid] = shape
+
+        log.debug('Completed loading ROIs and Shapes in Dataset: %s'
+                  % self.target_object.id.val)
+
 
 class ProjectWrapper(PDIWrapper):
 
@@ -894,6 +961,7 @@ class ImageWrapper(ValueWrapper):
     def __init__(self, value_resolver):
         super(ImageWrapper, self).__init__(value_resolver)
         self.rois_by_id = dict()
+        self.shapes_by_id = dict()
         self.rois_by_name = dict()
         self.ambiguous_naming = False
         self._load()
@@ -904,15 +972,25 @@ class ImageWrapper(ValueWrapper):
     def get_roi_name_by_id(self, rid):
         return unwrap(self.rois_by_id[rid].name)
 
+    def resolve_shape(self, value):
+        try:
+            return self.shapes_by_id[int(value)].id.val
+        except KeyError:
+            log.warn('Image is missing Shape: %s' % value)
+            return -1
+        except ValueError:
+            log.warn('Wrong input type for Shape ID: %s' % value)
+            return -1
+
     def resolve_roi(self, column, row, value):
         try:
             return self.rois_by_id[int(value)].id.val
         except KeyError:
             log.warn('Image is missing ROI: %s' % value)
-            return Skip()
+            return -1
         except ValueError:
             log.warn('Wrong input type for ROI ID: %s' % value)
-            return Skip()
+            return -1
 
     def _load(self):
         query_service = self.client.getSession().getQueryService()
@@ -930,9 +1008,10 @@ class ImageWrapper(ValueWrapper):
         while True:
             parameters.page(len(data), 1000)
             rv = query_service.findAllByQuery((
-                'select distinct r from Image as i '
-                'join i.rois as r '
-                'where i.id = :id order by r.id desc'),
+                'select distinct s from Shape as s '
+                'join s.roi as r '
+                'join r.image as i '
+                'where i.id = :id order by s.id desc'),
                 parameters, {'omero.group': '-1'})
             if len(rv) == 0:
                 break
@@ -943,15 +1022,19 @@ class ImageWrapper(ValueWrapper):
 
         rois_by_id = dict()
         rois_by_name = dict()
-        for roi in data:
+        shapes_by_id = dict()
+        for shape in data:
+            roi = shape.roi
             rid = roi.id.val
             rois_by_id[rid] = roi
+            shapes_by_id[shape.id.val] = shape
             if unwrap(roi.name) in rois_by_name.keys():
                 log.warn('Conflicting ROI names.')
                 self.ambiguous_naming = True
             rois_by_name[unwrap(roi.name)] = roi
         self.rois_by_id = rois_by_id
         self.rois_by_name = rois_by_name
+        self.shapes_by_id = shapes_by_id
         log.debug('Completed parsing image: %s' % self.target_name)
 
 
@@ -1148,8 +1231,8 @@ class ParsingContext(object):
                     if isinstance(value, basestring):
                         column.size = max(
                             column.size, len(value.encode('utf-8')))
-                    # The following are needed for
-                    # getting post process column sizes
+                    # The following IDs are needed for
+                    # post_process() to get column sizes for names
                     if column.__class__ is WellColumn:
                         column.values.append(value)
                     elif column.__class__ is ImageColumn:
@@ -1164,6 +1247,8 @@ class ParsingContext(object):
                     log.error('Original value "%s" now "%s" of bad type!' % (
                         original_value, value))
                     raise
+            # we call post_process on each single (mostly empty) row
+            # to get ids -> names
             self.post_process()
             for column in self.columns:
                 column.values = []
